# ruff: noqa: E501
import pytest

from hatch_ci import code, tools

# this is the output from ${{ toJson(github) }}
GITHUB = {
    "beta": {
        "ref": "refs/heads/beta/0.3.10",
        "sha": "507c657056d1a66520ec6b219a64706e70b0ff15",
        "run_number": "98",
        "run_id": "5904313530",
    },
    "release": {
        "ref": "refs/tags/release/0.3.9",
        "sha": "2b79a3669cd06e21741f6ee7a271903482344e76",
        "run_number": "22",
        "run_id": "5789585760",
    },
    "master": {
        "ref": "refs/heads/master",
        "sha": "9df889992f422a62cdd3ced41323eaba6e855ae3",
        "run_number": "257",
        "run_id": "5789314202",
    },
}


def splitter(txt):
    result = {}
    for line in txt.split("\n"):
        if not line.strip() or line.strip().startswith("#"):
            continue
        if len(kv := line.split("=")) == 2:
            result[kv[0].strip()] = kv[1].strip()
    return result


#
#
# def T(txt):
#     from textwrap import dedent
#
#     txt = dedent(txt)
#     if txt.startswith("\n"):
#         txt = txt[1:]
#     return txt
#
#
# def T1(txt):
#     return T(txt).rstrip("\n")


def test_list_of_paths():
    from pathlib import Path

    assert tools.list_of_paths([]) == []
    assert tools.list_of_paths("hello") == [Path("hello")]
    assert tools.list_of_paths(["hello", Path("world")]) == [
        Path("hello"),
        Path("world"),
    ]


def test_replace():
    fixers = {
        "abc": "def",
    }
    assert tools.replace("abcdef abc123", fixers) == "defdef def123"
    fixers = {
        "re:([ab])cde": "x\\1",
    }
    assert tools.replace("acde bcde123", fixers) == "xa xb123"

    fixers = {
        # for the github actions
        "re:(https://github.com/.+/actions/workflows/)(master)(.yml/badge.svg)": "\\1{{ ctx.workflow }}\\3",
        "re:(https://github.com/.+/actions)/(workflows/)(master.yml)(?!/)": "\\1/runs/{{ ctx.runid }}",
        # for the codecov part
        "re:(https://codecov.io/gh/.+/tree)/master(/graph/badge.svg[?]token=.+)": "\\1/{{ ctx.branch|urlquote }}\\2",
        "re:(https://codecov.io/gh/.+/tree)/master(?!/)": "\\1/{{ ctx.branch|urlquote }}",
    }

    txt = "https://github.com/cav71/setuptools-github/actions/workflows/master.yml/badge.svg"
    expected = "https://github.com/cav71/setuptools-github/actions/workflows/{{ ctx.workflow }}.yml/badge.svg"
    assert tools.replace(txt, fixers) == expected

    txt = "https://github.com/cav71/setuptools-github/actions/workflows/master.yml"
    expected = "https://github.com/cav71/setuptools-github/actions/runs/{{ ctx.runid }}"
    assert tools.replace(txt, fixers) == expected

    txt = "https://codecov.io/gh/cav71/setuptools-github/tree/master/graph/badge.svg?token=RANDOM123"
    expected = "https://codecov.io/gh/cav71/setuptools-github/tree/{{ ctx.branch|urlquote }}/graph/badge.svg?token=RANDOM123"
    assert tools.replace(txt, fixers) == expected

    txt = "https://codecov.io/gh/cav71/setuptools-github/tree/master"
    expected = (
        "https://codecov.io/gh/cav71/setuptools-github/tree/{{ ctx.branch|urlquote }}"
    )
    assert tools.replace(txt, fixers) == expected


def test_bump_version():
    "bump version test"
    assert tools.bump_version("0.0.1", "micro") == "0.0.2"
    assert tools.bump_version("0.0.2", "micro") == "0.0.3"
    assert tools.bump_version("0.0.2", "minor") == "0.1.0"
    assert tools.bump_version("1.2.3", "major") == "2.0.0"
    assert tools.bump_version("1.2.3", "release") == "1.2.3"


def test_update_version_master(git_project_factory):
    "test the update_version processing on the master branch"

    repo = git_project_factory().create("1.2.3")
    assert code.get_module_var(repo.initfile) == "1.2.3"

    # verify nothing has changed
    assert "1.2.3" == tools.update_version(repo.initfile, abort=False)
    assert code.get_module_var(repo.initfile) == "1.2.3"
    assert (
        code.get_module_var(repo.initfile, "__hash__")
        == repo(["rev-parse", "HEAD"])[:7]
    )

    assert "1.2.3" == tools.update_version(repo.initfile, GITHUB["master"], abort=False)
    assert code.get_module_var(repo.initfile) == "1.2.3"
    assert code.get_module_var(repo.initfile, "__hash__") == GITHUB["master"]["sha"][:7]


def test_update_version_beta(git_project_factory):
    "test the update_version processing on the master branch"

    repo = git_project_factory().create("0.3.10")
    assert code.get_module_var(repo.initfile) == "0.3.10"
    assert repo.branch() == "master"

    # branch
    repo.branch("beta/0.3.10", "master")
    assert repo.branch() == "beta/0.3.10"

    assert tools.update_version(repo.initfile, abort=False)
    assert code.get_module_var(repo.initfile) == "0.3.10b0"
    assert (
        code.get_module_var(repo.initfile, "__hash__")
        == repo(["rev-parse", "HEAD"])[:7]
    )
    repo.revert(repo.initfile)

    assert code.get_module_var(repo.initfile) == "0.3.10"
    assert tools.update_version(repo.initfile, GITHUB["beta"], abort=False)
    assert code.get_module_var(repo.initfile) == "0.3.10b98"
    assert code.get_module_var(repo.initfile, "__hash__") == "507c657"
    repo.revert(repo.initfile)

    # wrong branch
    repo.branch("beta/0.0.2", "master")
    assert repo.branch() == "beta/0.0.2"
    pytest.raises(
        tools.InvalidVersionError, tools.update_version, repo.initfile, abort=False
    )

    github_dump = GITHUB["beta"].copy()
    github_dump["ref"] = "refs/heads/beta/0.0.2"
    pytest.raises(
        tools.InvalidVersionError,
        tools.update_version,
        repo.initfile,
        github_dump,
        abort=False,
    )


def test_update_version_release(git_project_factory):
    repo = git_project_factory().create("0.3.9")
    assert code.get_module_var(repo.initfile) == "0.3.9"

    # branch
    repo.branch("beta/0.3.9", "master")
    assert repo.branch() == "beta/0.3.9"

    path = repo.workdir / "hello.txt"
    path.write_text("hello world\n")
    repo.commit(path, "initial")

    repo(["tag", "release/0.3.9", repo(["rev-parse", "HEAD"])[:7]])

    assert (
        tools.update_version(repo.initfile, GITHUB["release"], abort=False) == "0.3.9"
    )
    assert code.get_module_var(repo.initfile) == "0.3.9"
    assert (
        code.get_module_var(repo.initfile, "__hash__") == GITHUB["release"]["sha"][:7]
    )
    repo.revert(repo.initfile)


def test_get_environment(git_project_factory):
    repo = git_project_factory().create("0.3.9")
    assert repo.branch() == "master"
    assert code.get_module_var(repo.initfile) == "0.3.9"

    env = tools.get_environment(repo.initfile, None, None)
    out = (
        env.from_string("""
{% for key, value in sorted(ctx.items()) -%}
{{key}} = {{value}}
{% endfor %}
""")
        .render()
        .strip()
    )

    found = splitter(out)
    assert found == {
        "branch": "master",
        "build": "0",
        "current": "0.3.9",
        "ref": "refs/heads/master",
        "runid": "0",
        "sha": repo.head.target.hex,
        "version": "0.3.9",
        "workflow": "master",
    }

    repo.branch("beta/0.3.9", "master")
    assert repo.branch() == "beta/0.3.9"


def test_generate_build_record(tmp_path):
    dst = tmp_path / "abc.def"

    tools.generate_build_record(dst, {"a": 1, "b": 2})
    assert splitter(dst.read_text()) == {"a": "1", "b": "2"}


# def test_process(git_project_factory):
#
#     # generate the project
#     repo = git_project_factory().create("0.3.10")
#     assert repo.initfile.read_text() == '__version__ = "0.3.10"\n'
#
#     record = repo.initfile.parent / "_build.py"
#     assert not record.exists()
#
#     # we generate the build using a GITHUB_DUMP variable
#     tools.process(repo.initfile, json.dumps(GITHUB["beta"]), record)
#
#     assert repo.initfile.read_text() == f"""
# __version__ = "0.3.10b98"
# __hash__ = "{GITHUB['beta']['sha'][:7]}"
# """.strip()
#
#     assert record.read_text() == """
# # autogenerate build file
# branch = 'beta/0.3.10'
# build = '98'
# current = '0.3.10'
# ref = 'refs/heads/beta/0.3.10'
# runid = '5904313530'
# sha = '507c657056d1a66520ec6b219a64706e70b0ff15'
# version = '0.3.10b98'
# workflow = 'beta'
# """.lstrip()
#
#
# def test_process_fixers(git_project_factory):
#     def write_tfile(tfile):
#         tfile.write_text(
#             """
# {% for k, v in ctx.items() | sort -%}
# Key[{{k}}] = {{v}}
# {% endfor %}
# """
#         )
#         return tfile
#
#     repo = git_project_factory().create("1.2.3")
#
#     # tfile won't appear in the repo.status() because is untracked
#     tfile = write_tfile(repo.workdir / "test.txt")
#
#     data = tools.process(repo.initfile, None, "_build.py", tfile)
#
#     assert data["sha"][-1] != "*"
#
#     assert (
#         tfile.read_text()
#         == f"""
# Key[branch] = master
# Key[build] = 0
# Key[current] = 1.2.3
# Key[ref] = {data['ref']}
# Key[runid] = 0
# Key[sha] = {data['sha']}
# Key[version] = 1.2.3
# Key[workflow] = master
# """
#     )
#
#     # clean and switch to new branch
#     repo.revert(repo.initfile)
#     (repo.initfile.parent / "_build.py").unlink()
#
#     write_tfile(tfile)
#     repo.branch("beta/1.2.3", "master")
#
#     data = tools.process(repo.initfile, None, "_build.py", tfile)
#     assert data["sha"][-1] != "*"
#
#     assert (
#         tfile.read_text()
#         == f"""
# Key[branch] = beta/1.2.3
# Key[build] = 0
# Key[current] = 1.2.3
# Key[ref] = {data['ref']}
# Key[runid] = 0
# Key[sha] = {data['sha']}
# Key[version] = 1.2.3b0
# Key[workflow] = beta
# """
#     )



from __future__ import annotations

import ast
import re
from pathlib import Path
from typing import Any


class ValidationError(Exception):
    pass


class MissingVariableError(ValidationError):
    pass


def get_module_var(
    path: Path | str, var: str = "__version__", abort=True
) -> str | None:
    """extract from a python module in path the module level <var> variable

    Args:
        path (str,Path): python module file to parse using ast (no code-execution)
        var (str): module level variable name to extract
        abort (bool): raise MissingVariable if var is not present

    Returns:
        None or str: the variable value if found or None

    Raises:
        MissingVariable: if the var is not found and abort is True

    Notes:
        this uses ast to parse path, so it doesn't load the module
    """

    class V(ast.NodeVisitor):
        def __init__(self, keys):
            self.keys = keys
            self.result = {}

        def visit_Module(self, node):  # noqa: N802
            # we extract the module level variables
            for subnode in ast.iter_child_nodes(node):
                if not isinstance(subnode, ast.Assign):
                    continue
                for target in subnode.targets:
                    if target.id not in self.keys:
                        continue
                    if not isinstance(subnode.value, (ast.Num, ast.Str, ast.Constant)):
                        raise ValidationError(
                            f"cannot extract non Constant variable "
                            f"{target.id} ({type(subnode.value)})"
                        )
                    if isinstance(subnode.value, ast.Str):
                        value = subnode.value.s
                    elif isinstance(subnode.value, ast.Num):
                        value = subnode.value.n
                    else:
                        value = subnode.value.value
                    if target.id in self.result:
                        raise ValidationError(
                            f"found multiple repeated variables {target.id}"
                        )
                    self.result[target.id] = value
            return self.generic_visit(node)

    v = V({var})
    path = Path(path)
    if path.exists():
        tree = ast.parse(Path(path).read_text())
        v.visit(tree)
    if var not in v.result and abort:
        raise MissingVariableError(f"cannot find {var} in {path}", path, var)
    return v.result.get(var, None)


def set_module_var(
    path: str | Path, var: str, value: Any, create: bool = True
) -> tuple[Any, str]:
    """replace var in path with value

    Args:
        path (str,Path): python module file to parse
        var (str): module level variable name to extract
        value (None or Any): if not None replace var in version_file
        create (bool): create path if not present

    Returns:
        (str, str) the (<previous-var-value|None>, <the new text>)
    """

    # validate the var
    get_module_var(path, var, abort=False)

    # module level var
    expr = re.compile(f"^{var}\\s*=\\s*['\\\"](?P<value>[^\\\"']*)['\\\"]")
    fixed = None
    lines = []

    src = Path(path)
    if not src.exists() and create:
        src.parent.mkdir(parents=True, exist_ok=True)
        src.touch()

    input_lines = src.read_text().split("\n")
    for line in input_lines:
        if fixed is not None:
            lines.append(line)
            continue
        match = expr.search(line)
        if match:
            fixed = match.group("value")
            if value is not None:
                x, y = match.span(1)
                line = line[:x] + value + line[y:]
        lines.append(line)
    txt = "\n".join(lines)
    if (fixed is None) and create:
        if txt and txt[-1] != "\n":
            txt += "\n"
        txt += f'{var} = "{value}"'

    with Path(path).open("w") as fp:
        fp.write(txt)
    return fixed, txt



# see https://pypi.org/project/setuptools-github
# copy of setuptools_github.tools
from __future__ import annotations

import argparse
import functools
import json
import logging
import re
from pathlib import Path
from typing import Any

import jinja2

from . import code, common, fileos, scm, text

log = logging.getLogger(__name__)


class ToolsError(Exception):
    pass


class ValidationError(ToolsError):
    pass


class InvalidVersionError(ToolsError):
    pass


class MissingVariableError(ToolsError):
    pass


class _NO:
    pass


class _Context(argparse.Namespace):
    def items(self):
        for name, value in self.__dict__.items():
            if name.startswith("_"):
                continue
            yield (name, value)


def list_of_paths(paths: str | Path | list[str | Path] | None) -> list[Path]:
    if not paths:
        return []
    return [Path(s) for s in ([paths] if isinstance(paths, (str, Path)) else paths)]


backup = functools.partial(fileos.backup, ext=common.BACKUP_SUFFIX)
unbackup = functools.partial(fileos.unbackup, ext=common.BACKUP_SUFFIX)


def get_option(
    config: dict[str, str], var: str, typ: Any = _NO, fallback: Any = _NO
) -> Any:
    """extract an option value from the hatch config dict

    Eg.
        get_options(self.config, "key")
    """
    value = config.get(var, fallback)
    if value is _NO:
        raise ValidationError(f"cannot find variable '{var}' for plugin 'ci'")
    try:
        new_value = typ(value) if typ is not _NO else value
    except Exception as exc:
        raise ValidationError(f"cannot convert to {typ=} the {value=}") from exc
    return new_value


def replace(txt: str, replacements: dict[str, str] | None = None) -> str:
    result = txt
    for src, dst in (replacements or {}).items():
        if src.startswith("re:"):
            result = re.sub(src[3:], dst, result)
        else:
            result = result.replace(src, dst)
    return result


def bump_version(version: str, mode: str) -> str:
    """given a version str will bump it according to mode

    Arguments:
        version: text in the N.M.O form
        mode: major, minor or micro

    Returns:
        increased text

    >>> bump_version("1.0.3", "micro")
    "1.0.4"
    >>> bump_version("1.0.3", "minor")
    "1.1.0"
    """
    newver = [int(n) for n in version.split(".")]
    if mode == "major":
        newver[-3] += 1
        newver[-2] = 0
        newver[-1] = 0
    elif mode == "minor":
        newver[-2] += 1
        newver[-1] = 0
    elif mode == "micro":
        newver[-1] += 1
    return ".".join(str(v) for v in newver)


def _validate_gdata(
    gdata: dict[str, Any], abort: bool = True
) -> tuple[set[str], set[str]]:
    keys = {
        "ref",
        "sha",
        "run_id",
        "run_number",
    }
    missing = keys - set(gdata)
    extra = set(gdata) - keys
    if abort and missing:
        raise ToolsError(
            f"missing keys from gdata '{','.join(missing)}'", missing, extra, gdata
        )
    return missing, extra


def get_data(
    version_file: str | Path,
    github_dump: str | None = None,
    record_path: Path | None = None,
    abort: bool = True,
) -> dict[str, str | None]:
    """extracts version information from github_dump and updates version_file in-place

    This gather version, branch, commit and other info from:

    - version_file (eg. the __init__.py file containing __version__)
    - github_dump *a json encoded dictionary in GITHUB_DUMP
    - a record file (eg. _build.py)
    - the git checkout (using git)

    Args:
        version_file (str, Path): path to a file  with a __version__ variable
        github_dump (str): the os.getenv("GITHUB_DUMP") value (a json structure)
        record: pull data from a _build.py file

    Returns:
        dict[str,str|None]: a dict with the current config
        dict[str,str|None]: a dict with the github dump data

    Example:
        for github data:
            {
                "ref": "refs/heads/beta/0.3.10",
                "run_id": "5904313530",
                "run_number": "98",
                "sha": "507c657056d1a66520ec6b219a64706e70b0ff15",
            }
        for data:
            {
                "branch": "beta/0.3.10",
                "build": "98",
                "current": "0.3.10",
                "ref": "refs/heads/beta/0.3.10",
                "runid": "5904313530",
                "sha": "507c657056d1a66520ec6b219a64706e70b0ff15",
                "version": "0.3.10b98",
                "workflow": "beta",
            }
    """

    # we fill this structure getting data from these sources:
    #  github_dump -> record_path (eg. _build.py) -> repo infos (eg. the git checkout)
    data = {
        "version": code.get_module_var(version_file, "__version__"),
        "current": code.get_module_var(version_file, "__version__"),
        "ref": None,
        "branch": None,
        "sha": None,
        "build": None,
        "runid": None,
        "workflow": None,
    }

    path = Path(version_file)
    repo = scm.lookup(path)
    record = record_path.exists() if record_path else None

    if not (repo or github_dump or record):
        if abort:
            raise scm.InvalidGitRepoError(
                f"cannot figure out settings (no repo in {path}, "
                f"a GITHUB_DUMP or a _build.py file)"
            )
        return data

    dirty = False
    if github_dump:
        gdata = json.loads(github_dump) if isinstance(github_dump, str) else github_dump
    elif record_path and record_path.exists():
        mod = fileos.loadmod(record_path)
        gdata = {
            "ref": mod.ref,
            "sha": mod.sha,
            "run_number": mod.build,
            "run_id": mod.runid,
        }
    elif repo:
        gdata = {
            "ref": repo.head.name,
            "sha": repo.head.target.hex,
            "run_number": 0,
            "run_id": 0,
        }
        dirty = repo.dirty()
    else:
        raise RuntimeError(
            "Unable to retrieve data from GITHUB_DUMP or " "{record_path=} or {repo=}"
        )

    # make sure we have all keys
    _validate_gdata(gdata)

    expr = re.compile(r"/(?P<what>beta|release)/(?P<version>\d+([.]\d+)*)$")
    expr1 = re.compile(r"(?P<version>\d+([.]\d+)*)(?P<num>b\d+)?$")

    # update/fill the data values
    data["ref"] = gdata["ref"]
    data["sha"] = gdata["sha"] + ("*" if dirty else "")
    data["build"] = gdata["run_number"]  # unless in GITHUB_DUMP these are likely None
    data["runid"] = gdata["run_id"]  # unless in GITHUB_DUMP these are likely None

    data["branch"] = text.lstrip(gdata["ref"], ["refs/heads/", "refs/tags/"])
    data["workflow"] = data["branch"]

    current = data["current"]
    if match := expr.search(gdata["ref"]):
        # setuptools double calls the update_version,
        # this fixes the issue
        match1 = expr1.search(current or "")
        if not match1:
            raise InvalidVersionError(f"cannot parse current version '{current}'")
        if match1.group("version") != match.group("version"):
            raise InvalidVersionError(
                f"building package for {current} from '{gdata['ref']}' "
                f"branch ({match.groupdict()} mismatch {match1.groupdict()})"
            )
        if match.group("what") == "beta":
            data["version"] = f"{match1.group('version')}b{gdata['run_number']}"
            data["workflow"] = "beta"
        else:
            data["workflow"] = "tags"

    assert len(data) == 8, "cannot change the data structure shape"  # noqa: S101
    return data


def update_version(
    version_file: str | Path, github_dump: str | None = None, abort: bool = True
) -> str | None:
    """extracts version information from github_dump and updates version_file in-place

    Args:
        version_file (str, Path): path to a file with a __version__ variable
        github_dump (str): the os.getenv("GITHUB_DUMP") value

    Returns:
        str: the new version for the package
    """

    data = get_data(version_file, github_dump, abort=abort)
    code.set_module_var(version_file, "__version__", data["version"])
    code.set_module_var(version_file, "__hash__", (data["sha"] or "")[:7])
    return data["version"]


def get_environment(
    version_file: str | Path,
    github_dump: str | None = None,
    record_path: Path | None = None,
    abort: bool = True,
) -> jinja2.Environment:
    """returns a context object"""

    from urllib.parse import quote

    env = jinja2.Environment(autoescape=True)
    env.filters["urlquote"] = functools.partial(quote, safe="")
    env.globals = {
        "dir": dir,
        "len": len,
        "sorted": sorted,
        "reversed": reversed,
    }
    data = get_data(version_file, github_dump, record_path, abort)
    env.globals["ctx"] = _Context(**data)
    return env


def generate_build_record(record_path: Path, data: dict[str, Any]) -> Path:
    record_path.parent.mkdir(parents=True, exist_ok=True)
    with record_path.open("w") as fp:
        print("# autogenerate build file", file=fp)
        for key, value in sorted((data or {}).items()):
            value = f"'{value}'" if isinstance(value, str) else value
            print(f"{key} = {value}", file=fp)
    return record_path


# def process(
#     version_file: str | Path,
#     github_dump: str | None = None,
#     record: str | Path = "_build.py",
#     paths: str | Path | list[str | Path] | None = None,
#     fixers: dict[str, str] | None = None,
#     backup: Callable[[Path | str], None] | None = None,
#     abort: bool = True,
# ) -> dict[str, str | None]:
#     """get version from github_dump and updates version_file/paths
#
#     Args:
#         version_file (str, Path): path to a file with __version__ variable
#         github_dump (str): the os.getenv("GITHUB_DUMP") value
#         paths (str, Path): path(s) to files jinja2 processeable
#         fixers (dict[str,str]): fixer dictionary
#         record: set to True will generate a _build.py sibling of version_file
#
#     Returns:
#         str: the new version for the package
#
#     Example:
#         {'branch': 'beta/0.3.1',
#          'build': 0,
#          'current': '0.3.1',
#          'hash': 'c9e484a*',
#          'version': '0.3.1b0',
#          'runid': 0
#         }
#     """
#     from argparse import Namespace
#     from functools import partial
#     from urllib.parse import quote
#
#     from jinja2 import Environment
#
#     class Context(Namespace):
#         def items(self):
#             for name, value in self.__dict__.items():
#                 if name.startswith("_"):
#                     continue
#                 yield (name, value)
#
#     record_path = (Path(version_file).parent / record).absolute() if record else None
#     data = get_data(version_file, github_dump, record_path, abort)
#
#     if not backup:
#
#         def backup(_: Path | str) -> None:
#             pass
#
#     breakpoint()
#     # backup(version_file)
#
#     # set_module_var(version_file, "__version__", data["version"])
#     # set_module_var(version_file, "__hash__", (data["sha"] or "")[:7])
#
#     # env = Environment(autoescape=True)
#     # env.filters["urlquote"] = partial(quote, safe="")
#     # for path in list_of_paths(paths):
#     #     txt = replace(path.read_text(), fixers)
#     #     tmpl = env.from_string(txt)
#     #     backup(path)
#     #     path.write_text(tmpl.render(ctx=Context(**data)))
#     #
#     # if record_path:
#     #     generate_build_record(record_path, data)
#
#     return data
#
#



from __future__ import annotations


def indent(txt: str, pre: str = " " * 2) -> str:
    """simple text indentation"""

    from textwrap import dedent

    txt = dedent(txt)
    if txt.endswith("\n"):
        last_eol = "\n"
        txt = txt[:-1]
    else:
        last_eol = ""

    result = pre + txt.replace("\n", "\n" + pre) + last_eol
    return result if result.strip() else result.strip()


def lstrip(txt: str, ending: str | list[str]) -> str:
    endings = ending if isinstance(ending, list) else [ending]
    for left in endings:
        txt = txt[len(left) :] if txt.startswith(left) else txt
    return txt



from hatch_ci import exceptions


def test_abort_exception():
    "test the AbortExecution exception"
    a = exceptions.AbortExecutionError(
        "a one-line error message",
        """
        A multi line
          explanation of
           what happened
         with some detail
    """,
        """
    Another multiline hint how
      to fix the issue
    """,
    )

    assert a.message == "a one-line error message"
    assert (
        f"\n{a.explain}\n"
        == """
A multi line
  explanation of
   what happened
 with some detail
"""
    )
    assert (
        f"\n{a.hint}\n"
        == """
Another multiline hint how
  to fix the issue
"""
    )

    assert (
        f"\n{a!s}\n"
        == """
a one-line error message
  A multi line
    explanation of
     what happened
   with some detail
hint:
  Another multiline hint how
    to fix the issue
"""
    )

    a = exceptions.AbortExecutionError("hello world")
    assert a.message == "hello world"
    assert a.explain == ""
    assert a.hint == ""
    assert str(a) == "hello world"



"""various file/dir related functions"""

from __future__ import annotations

import os
import types
from pathlib import Path
from typing import Any, overload


class FileOSError(Exception):
    pass


class FileOSModuleNotFoundError(FileOSError):
    pass


class FileOSMInvalidModuleError(FileOSError):
    pass


def rmtree(path: Path):
    """universal (win|*nix) rmtree"""

    from os import name
    from shutil import rmtree
    from stat import S_IWUSR

    if name == "nt":
        for p in path.rglob("*"):
            p.chmod(S_IWUSR)
    rmtree(path, ignore_errors=True)
    if path.exists():
        raise RuntimeError(f"cannot remove {path=}")


def mkdir(path: Path) -> Path:
    """make a path directory and returns if it has been created"""
    path.mkdir(exist_ok=True, parents=True)
    return path


def touch(path: Path) -> Path:
    """touch a new empty file"""
    mkdir(path.parent)
    path.write_text("")
    return path


@overload
def which(exe: Path | str, kind: type[list], abort: bool = True) -> list[Path]: ...


@overload
def which(exe: Path | str, kind: None, abort: bool = True) -> Path | None: ...


def which(
    exe: Path | str, kind: type[list] | None = None, abort: bool = True
) -> list[Path] | Path | None:
    candidates: list[Path] = []
    for srcdir in os.environ.get("PATH", "").split(os.pathsep):
        for ext in os.environ.get("PATHEXT", "").split(os.pathsep):
            path = srcdir / Path(exe).with_suffix(ext)
            if not path.exists():
                continue
            if kind is None:
                return path
            candidates.append(path)
    return candidates


def loadmod(path: Path | str, suffix: str | None = "") -> types.ModuleType:
    import inspect
    from importlib import machinery, util

    if isinstance(path, str):
        if not Path(path).is_absolute():
            path = Path(inspect.stack()[1].filename).parent / path
        path = Path(path)

    if suffix is not None:
        machinery.SOURCE_SUFFIXES.append(suffix)
    try:
        spec = util.spec_from_file_location(Path(path).name, Path(path))
        if not spec:
            raise FileOSModuleNotFoundError(f"cannot find module for {path=}")
        module = util.module_from_spec(spec)
        if not spec.loader:
            raise FileOSMInvalidModuleError(f"invalid module in {path=}")
        spec.loader.exec_module(module)
    finally:
        if suffix is not None:
            machinery.SOURCE_SUFFIXES.pop()
    return module


### FILE UTILITIES


def zextract(path: Path | str, items: list[str] | None = None) -> dict[str, Any]:
    """extracts from path (a zipfile/tarball) all data in a dictionary"""
    from tarfile import TarFile, is_tarfile
    from zipfile import ZipFile, is_zipfile

    path = Path(path)
    result = {}
    if is_tarfile(path):
        with TarFile.open(path) as tfp:
            for member in tfp.getmembers():
                fp = tfp.extractfile(member)
                if not fp:
                    continue
                result[member.name] = str(fp.read(), encoding="utf-8")
    elif is_zipfile(path):
        with ZipFile(path) as tfp:
            for zinfo in tfp.infolist():
                if items and zinfo.filename not in items:
                    continue
                with tfp.open(zinfo.filename) as fp:
                    result[zinfo.filename] = str(fp.read(), encoding="utf-8").replace(
                        "\r", ""
                    )

    return result


def backup(path: Path, ext: str, overwrite: bool = False, abort: bool = True) -> Path:
    """creates a backup of path"""
    from shutil import copyfile, copymode

    path2 = path.parent / f"{path.name}{ext}"
    if path2.exists() and not overwrite:
        if abort:
            raise FileOSError(f"backup file present {path2}")
        return path2

    copyfile(path, path2)
    copymode(path, path2)
    return path2


def unbackup(path: Path, ext: str, abort: bool = True) -> Path:
    """restores from a backup of path"""
    from shutil import move

    path2 = path.parent / f"{path.name}{ext}"
    if abort and not path2.exists():
        raise FileOSError(f"cannot find backup file {path2} for {path=}")
    if path2.exists():
        move(str(path2), str(path))
    return path2



import sys
from pathlib import Path

import pytest

from hatch_ci import fileos


def test_rmtree(tmp_path):
    target = tmp_path / "abc" / "def"
    target.mkdir(parents=True, exist_ok=True)
    assert target.exists()

    fileos.rmtree(target)
    assert not target.exists()
    assert target.parent.exists()


def test_mkdir(tmp_path):
    target = tmp_path / "abc"
    assert not target.exists()
    assert fileos.mkdir(target)
    assert target.exists()
    assert target.is_dir()


def test_touch(tmp_path):
    target = tmp_path / "abc"
    assert not target.exists()
    assert fileos.touch(target)
    assert target.exists()
    assert target.is_file()


def test_which():
    exe = "cmd" if sys.platform == "win32" else "sh"

    path = fileos.which(exe)
    assert path
    assert isinstance(path, Path)

    path = fileos.which(exe, kind=list)
    assert path
    assert isinstance(path, list)
    assert path[0] == fileos.which(exe)


def test_loadmod(tmp_path):
    path = fileos.touch(tmp_path / "blah")
    path.write_text("MYVAR = 99")

    mod = fileos.loadmod(path)
    assert mod.MYVAR == 99

    pytest.raises(FileNotFoundError, fileos.loadmod, tmp_path / "xyz")


def test_zextract(resolver):
    ball = resolver.lookup("foobar-0.0.0-py3-none-any.whl")
    data = fileos.zextract(ball)
    assert data["foobar/__init__.py"].strip() == '__version__ = "0.0.0"'

    ball = resolver.lookup("foobar-0.0.0.tar.gz")
    data = fileos.zextract(ball)
    assert (
        data["foobar-0.0.0/src/foobar/__init__.py"].strip() == '__version__ = "0.0.0"'
    )


def test_backup_unbackup(tmp_path):
    path = tmp_path / "anoter.test.txt"
    path.write_text("A brand new message")
    bak = path.parent / f"{path.name}.original"

    assert not bak.exists()

    assert fileos.backup(path, ".original") == bak
    assert bak.exists()
    pytest.raises(fileos.FileOSError, fileos.backup, path, ".original")
    path.write_text("New message")

    assert fileos.unbackup(path, ".original") == bak
    assert not bak.exists()
    pytest.raises(fileos.FileOSError, fileos.unbackup, path, ".original")
    assert path.read_text() == "A brand new message"



from hatchling.plugin import hookimpl

from hatch_ci.hook_build import CIBuildHook
from hatch_ci.hook_version import CIVersionSource


@hookimpl
def hatch_register_version_source():
    return CIVersionSource


@hookimpl
def hatch_register_build_hook():
    return CIBuildHook



import re
from pathlib import Path

import pytest
from build.__main__ import main as build

import hatch_ci
from hatch_ci import fileos, text

pytestmark = pytest.mark.skipif(True, reason="not ready yet")


def T(txt: str) -> str:  # noqa: N802
    return text.indent(txt, pre="").replace("\r", "").rstrip("\n")


@pytest.fixture(scope="function")
def project(git_project_factory, monkeypatch):
    """creates a project with this structure
    ├── .git/
    ├── pyproject.toml
    └── src/
        └── <name>>/
            ├── __init__.py  <- contains <version>
            └── xyz.py
    """

    def _make(name, version, cwd=True, use_hatch_ci_from_current_env=False):
        repo = git_project_factory(name=name).create(version)
        if cwd:
            monkeypatch.chdir(repo.workdir if cwd is True else cwd)

        paths = []

        # xyz module
        paths.append(repo.workdir / "src" / name / "template.jinja2")
        paths[-1].write_text(
            T("""
        # This is a file that will be jinjia2 processed during build
        # This string -> 'replace-me' will be replaced
        """)
        )

        # pyproject
        # in use_hatch_in_src we use the hatch-ci from the current src directory
        # (eg. pip editable installed)
        srcdir = str(Path(hatch_ci.__file__).parent.parent.parent).replace("\\", "/")
        hatch_src = "hatch-ci" if use_hatch_ci_from_current_env else f"-e file://{srcdir}"
        paths.append(repo.workdir / "pyproject.toml")
        paths[-1].write_text(f"""
[build-system]
requires = [
    "hatchling>=1.1.0",
    "typing-extensions",
    "jinja2",
    "{hatch_src}"
]
build-backend = "hatchling.build"

[project]
name = "{name}"
dynamic = ["version"]
description = "test project"
requires-python = ">= 3.8"
packages = ["src/{name}"]

[tool.hatch.version]
source = "ci"
version-file = "src/{name}/__init__.py"

[tool.hatch.build.hooks.ci]
version-file = "src/{name}/__init__.py"
process-replace = [
    ["re:(replace-me)", "[\\\\1 -> REPLACED!!]"]
]
process-paths = [
    "src/{name}/template.jinja2"
]
    """)
        repo.commit(paths, "init")
        return repo

    return _make


def match_version(txt):
    result = re.search(
        r"""
__version__ = "(?P<version>\d+([.]\d+)*)(?P<tag>[a-zA-Z].*)?"
__hash__ = "(?P<sha>[0-9abcdef]{7})"
""".strip(),
        txt.strip().replace("\r", ""),
    )
    return result.groupdict() if result else None


def match_build(txt):
    result = re.search(
        r"""
branch = '(?P<branch>[^']+)'
build = 0
current = '(?P<current>\d+([.]\d+)*)(?P<ctag>[a-zA-Z].*)?'
ref = '(?P<ref>[^']+)'
runid = 0
sha = '(?P<sha>[0-9abcdef]{40})'
version = '(?P<version>\d+([.]\d+)*)(?P<tag>[a-zA-Z].*)?'
workflow = '(?P<workflow>[^']+)'
    """.strip(),
        txt.strip().replace("\r", ""),
    )
    return result.groupdict() if result else None


def test_master_branch(project):
    """build sdist and wheel from master branch"""
    repo = project("foobar", "0.0.0", use_hatch_ci_from_current_env=True)
    assert repo.status() == {}

    # 1. check the jinja2 processed files content
    assert (repo.workdir / "src" / "foobar" / "template.jinja2").read_text() == T("""
    # This is a file that will be jinjia2 processed during build
    # This string -> 'replace-me' will be replaced
    """)

    # 2. build the package using this code
    build(["-n", "."], "pytest")

    # 3. verify we don't have leftovers after build
    assert repo.status() == {
        "dist/": 128,
    }

    # 4. verify the sdist contains the right files
    tarball = repo.workdir / "dist" / f"{repo.name}-{repo.version()}.tar.gz"
    contents = fileos.zextract(tarball)
    assert set(contents).difference({"foobar-0.0.0/.gitignore"}) == {
        "foobar-0.0.0/PKG-INFO",
        "foobar-0.0.0/pyproject.toml",
        "foobar-0.0.0/src/foobar/__init__.py",
        "foobar-0.0.0/src/foobar/__init__.py.original",
        "foobar-0.0.0/src/foobar/_build.py",
        "foobar-0.0.0/src/foobar/template.jinja2",
        "foobar-0.0.0/src/foobar/template.jinja2.original",
    }

    assert match_version(contents["foobar-0.0.0/src/foobar/__init__.py"]) == {
        "version": "0.0.0",
        "sha": repo.head.target.hex[:7],
        "tag": None,
    }
    assert match_build(contents["foobar-0.0.0/src/foobar/_build.py"]) == {
        "branch": "master",
        "current": "0.0.0",
        "sha": repo.head.target.hex,
        "version": "0.0.0",
        "tag": None,
        "workflow": "master",
        "ref": "refs/heads/master",
        "ctag": None,
    }
    assert T(contents["foobar-0.0.0/src/foobar/template.jinja2"]) == T("""
    # This is a file that will be jinjia2 processed during build
    # This string -> '[replace-me -> REPLACED!!]' will be replaced
    """)

    # 3. verify the wheel contains the right files
    wheel = repo.workdir / "dist" / f"{repo.name}-{repo.version()}-py3-none-any.whl"

    contents = fileos.zextract(wheel)
    assert set(contents) == {
        "foobar-0.0.0.dist-info/METADATA",
        "foobar-0.0.0.dist-info/RECORD",
        "foobar-0.0.0.dist-info/WHEEL",
        "foobar/__init__.py",
        "foobar/__init__.py.original",
        "foobar/_build.py",
        "foobar/template.jinja2",
        "foobar/template.jinja2.original",
    }

    assert match_version(contents["foobar/__init__.py"]) == {
        "version": "0.0.0",
        "sha": repo.head.target.hex[:7],
        "tag": None,
    }
    assert match_build(contents["foobar/_build.py"]) == {
        "branch": "master",
        "current": "0.0.0",
        "sha": repo.head.target.hex,
        "version": "0.0.0",
        "tag": None,
        "workflow": "master",
        "ref": "refs/heads/master",
        "ctag": None,
    }


def test_beta_branch(project):
    tag = "b0"  # we build with an index of 0 (the fallback)

    repo = project("foobar", "0.0.0", use_hatch_ci_from_current_env=True)
    repo.branch("beta/0.0.0")
    assert repo.branch() == "beta/0.0.0"

    assert repo.status() == {}
    build(["-n", "."], "pytest")

    # 1. verify we don't have leftovers
    assert repo.status() == {
        "dist/": 128,
    }

    # 2. verify the sdist contains the right files
    tarball = repo.workdir / "dist" / f"{repo.name}-{repo.version()}{tag}.tar.gz"

    contents = fileos.zextract(tarball)
    assert set(contents).difference({f"foobar-0.0.0{tag}/.gitignore"}) == {
        f"foobar-0.0.0{tag}/PKG-INFO",
        f"foobar-0.0.0{tag}/pyproject.toml",
        f"foobar-0.0.0{tag}/src/foobar/__init__.py",
        f"foobar-0.0.0{tag}/src/foobar/__init__.py.original",
        f"foobar-0.0.0{tag}/src/foobar/_build.py",
        f"foobar-0.0.0{tag}/src/foobar/template.jinja2",
        f"foobar-0.0.0{tag}/src/foobar/template.jinja2.original",
    }

    assert match_version(contents[f"foobar-0.0.0{tag}/src/foobar/__init__.py"]) == {
        "version": "0.0.0",
        "sha": repo.head.target.hex[:7],
        "tag": tag,
    }
    assert match_build(contents[f"foobar-0.0.0{tag}/src/foobar/_build.py"]) == {
        "branch": "beta/0.0.0",
        "current": "0.0.0",
        "sha": repo.head.target.hex,
        "version": "0.0.0",
        "tag": tag,
        "ref": "refs/heads/beta/0.0.0",
        "workflow": "beta",
        "ctag": None,
    }

    # 3. verify the wheel contains the right files
    wheel = (
        repo.workdir / "dist" / f"{repo.name}-{repo.version()}{tag}-py3-none-any.whl"
    )

    contents = fileos.zextract(wheel)
    assert set(contents) == {
        f"foobar-0.0.0{tag}.dist-info/METADATA",
        f"foobar-0.0.0{tag}.dist-info/RECORD",
        f"foobar-0.0.0{tag}.dist-info/WHEEL",
        "foobar/__init__.py",
        "foobar/__init__.py.original",
        "foobar/_build.py",
        "foobar/template.jinja2",
        "foobar/template.jinja2.original",
    }

    assert match_version(contents["foobar/__init__.py"]) == {
        "version": "0.0.0",
        "sha": repo.head.target.hex[:7],
        "tag": "b0",
    }
    assert match_build(contents["foobar/_build.py"]) == {
        "branch": "beta/0.0.0",
        "current": "0.0.0",
        "ref": "refs/heads/beta/0.0.0",
        "sha": repo.head.target.hex,
        "version": "0.0.0",
        "tag": "b0",
        "workflow": "beta",
        "ctag": tag,
    }



from hatch_ci import text


def test_indent():
    txt = """\
     An unusually complicated text
    with un-even indented lines
   that make life harder
"""
    assert (
        text.indent(txt, pre="..")
        == """\
..  An unusually complicated text
.. with un-even indented lines
..that make life harder
"""
    )


def test_indent_another():
    txt = """
    This is a simply
       indented text
      with some special
         formatting
"""
    expected = """
..This is a simply
..   indented text
..  with some special
..     formatting
"""

    found = text.indent(txt[1:], "..")
    assert f"\n{found}" == expected


def test_lstrip():
    assert text.lstrip("/a/b/c/d/e", "/a/b") == "/c/d/e"



from __future__ import annotations

from . import text


class AbortExecutionError(Exception):
    @staticmethod
    def _strip(txt):
        txt = txt or ""
        txt = txt[1:] if txt.startswith("\n") else txt
        txt = text.indent(txt, pre="")
        return txt[:-1] if txt.endswith("\n") else txt

    def __init__(
        self, message: str, explain: str | None = None, hint: str | None = None
    ):
        self.message = message.strip()
        self._explain = explain
        self._hint = hint

    @property
    def explain(self):
        return self._strip(self._explain)

    @property
    def hint(self):
        return self._strip(self._hint)

    def __str__(self):
        result = [self.message]
        if self.explain:
            result.append(text.indent("\n" + self.explain, pre=" " * 2)[2:])
        if self.hint:
            result.extend(["\nhint:", text.indent("\n" + self.hint, pre=" " * 2)[2:]])
        return "".join(result)



from __future__ import annotations

from hatchling.version.source.plugin.interface import VersionSourceInterface

from . import common


class ValidationError(Exception):
    pass


class CIVersionSource(VersionSourceInterface):
    PLUGIN_NAME = common.PLUGIN_NAME

    def get_version_data(self):
        from os import getenv
        from pathlib import Path

        from hatch_ci import tools

        # paths = extract(self.config, "paths", typ=tools.list_of_paths, fallback=[])
        # fixers = extract(self.config, "fixers", typ=get_fixers, fallback={})

        version_file = Path(self.root) / tools.get_option(self.config, "version-file")
        record_path = (Path(version_file).parent / common.RECORD_NAME).absolute()

        if not version_file.exists():
            raise ValidationError(
                f"no 'version-file' key foun in plugin {self.PLUGIN_NAME}"
            )

        gdata = tools.get_data(version_file, getenv("GITHUB_DUMP"), record_path)

        # gdata = tools.process(
        #    version_file, getenv("GITHUB_DUMP"), paths=paths, fixers=fixers
        # )
        return {"version": gdata["version"]}



# This file is not mean to be used!!!
# ONLY FOR INTERNAL DEBUG

# wraps each method in a derived class of:
# - VersionSourceInterface
# - BuildHookInterface
# etc.

# wraps this:
# site-packages\pyproject_hooks\_impl.py ->
# def _call_hook(self, hook_name, kwargs):
#   ...
#   from hatch_ci._support import tracer
#   with tracer.decorator(_in_proc_script_path)() as script:

# change the OUTPUT below
from __future__ import annotations

import functools
import inspect
import io
import json
import os
import sys
from pathlib import Path
from typing import Any

OUTPUT = None  # eg. r"C:\Users\antonio\Projects\github\hatch-ci\NOTES.txt"


def is_scm(path: Path) -> str:
    return "<git>" if (Path(path) / ".git").exists() else "<plain>"


def jformat(data: Any, pre: str) -> str:
    return json.dumps(data, indent=2, sort_keys=True).replace(
        "\n", f"\n{' ' * len(pre)}"
    )


class Tracer:
    def __init__(self, path: Path | str | None = None):
        self.path = Path(path) if path else None

    def decorator(self, fn):
        if fn.__name__ == "_in_proc_script_path":
            return self.wraps__in_proc_script_path(fn)
        else:
            return self.wraps_interface(fn)

    def append(self, txt):
        if not self.path:
            return
        with self.path.open("a") as fp:
            fp.write(txt.getvalue() if hasattr(txt, "getvalue") else txt)

    def wraps__in_proc_script_path(self, fn):
        # this is mean to wrap the _in_proc_script_path function
        @functools.wraps(fn)
        def _fn(*args, **kwargs):
            # ok, looking to wrap _call_hook in _impl.py
            frames = [
                f
                for f in inspect.getouterframes(inspect.currentframe())
                if "_impl.py" in f.filename and f.function == "_call_hook"
            ]
            if len(frames) != 1:
                raise RuntimeError("cannot find _call_hook in _impl.py frame")
            variables = frames[0].frame.f_locals

            if not self.path:
                fp = io.StringIO()
                print("", file=fp)
                print(f"[pid={os.getpid()}]", file=fp)
                print(f"<hook> '{variables['hook_name']}'", file=fp)
                scm = (
                    "<git>"
                    if (Path(variables["self"].source_dir) / ".git").exists()
                    else "<plain>"
                )
                print(
                    f"  cwd (self.source_dir): {variables['self'].source_dir} {scm}",
                    file=fp,
                )
                print(f"  kwargs: {variables['kwargs']}", file=fp)
                print(f"  extra_environ: {variables['extra_environ']}", file=fp)
                print("  input.json:", file=fp)
                pre = "  input.json: "
                print(f"{pre}{jformat(variables['hook_input'], pre=pre)}", file=fp)
                self.append(fp)

            return fn(*args, **kwargs)

        return _fn

    def wraps_interface(self, method):
        # a method decorator
        # you should wrap methods of subclasses of:
        # - VersionSourceInterface
        # - BuildHookInterface
        # etc.
        @functools.wraps(method)
        def _fn(slf, *args, **kwargs):
            if self.path:
                fp = io.StringIO()
                print("", file=fp)
                print(f"[pid={os.getpid()}] {sys.argv=}", file=fp)
                print(
                    f"!callback! '{method.__name__}' {slf.__class__.__name__}", file=fp
                )
                print(f"  self.root: {slf.root} {is_scm(slf.root)}", file=fp)
                print(f"  config: {jformat(slf.config, '  config: ')}", file=fp)
                self.append(fp)
            return method(slf, *args, **kwargs)

        return _fn if self.path else method


tracer = Tracer()  # r"C:\Users\antonio\Projects\github\hatch-ci\NOTES.txt")



import os
from pathlib import Path

from hatchling.builders.hooks.plugin.interface import BuildHookInterface

from . import common


class CIBuildHook(BuildHookInterface):
    PLUGIN_NAME = common.PLUGIN_NAME

    def initialize(self, version, build_data):
        from . import code, tools

        version_file = Path(self.root) / tools.get_option(self.config, "version-file")
        record_path = (Path(version_file).parent / common.RECORD_NAME).absolute()
        env = tools.get_environment(version_file, os.getenv("GITHUB_DUMP"), record_path)

        tools.generate_build_record(record_path, env.globals["ctx"])

        tools.backup(version_file, abort=False)
        code.set_module_var(version_file, "__version__", env.globals["ctx"].version)
        code.set_module_var(
            version_file, "__hash__", (env.globals["ctx"].sha or "")[:7]
        )

        paths = [
            Path(self.root) / path
            for path in tools.get_option(self.config, "process-paths", fallback=[])
        ]
        replacements = tools.get_option(
            self.config, "process-replace", fallback=[], typ=dict
        )
        for path in paths:
            txt = tools.replace(path.read_text(), replacements)
            out = env.from_string(txt).render()
            tools.backup(path, abort=False)
            path.write_text(out)

    def finalize(self, version, build_data, artifact_path):
        from . import tools

        version_file = Path(self.root) / tools.get_option(self.config, "version-file")
        record_path = (Path(version_file).parent / common.RECORD_NAME).absolute()

        paths = [
            Path(self.root) / path
            for path in tools.get_option(self.config, "process-paths", fallback=[])
        ]
        for path in paths:
            tools.unbackup(path)
        tools.unbackup(version_file, abort=False)
        record_path.unlink()


if __name__ == "__main__":
    import toml

    from . import tools

    cfg = toml.loads(Path("pyproject.toml").read_text())
    replacements = dict(cfg["tool"]["hatch"]["build"]["hooks"]["ci"]["process-replace"])
    env = tools.get_environment(Path("src/hatch_ci/__init__.py"))

    path = Path("README.md")
    txt = tools.replace(path.read_text(), replacements)
    out = env.from_string(txt).render()
    Path("out.txt").write_text(out)



import pytest

from hatch_ci import code


def test_get_module_var(tmp_path):
    "pulls variables from a file"
    path = tmp_path / "in0.txt"
    path.write_text(
        """
# a test file
A = 12
B = 3+5
C = "hello"
# end of test
"""
    )
    assert 12 == code.get_module_var(path, "A")
    assert "hello" == code.get_module_var(path, "C")
    pytest.raises(code.ValidationError, code.get_module_var, path, "B")
    pytest.raises(code.MissingVariableError, code.get_module_var, path, "X1")

    path.write_text(
        """
# a test file
A = 12
B = 3+5
C = "hello"
C = "hello2"
# end of test
"""
    )
    pytest.raises(code.ValidationError, code.get_module_var, path, "C")


def test_set_module_var(tmp_path):
    "handles set_module_var cases"
    path = tmp_path / "in2.txt"

    path.write_text(
        """
# a fist comment line
__hash__ = "4.5.6"
# end of test
"""
    )

    version, txt = code.set_module_var(path, "__version__", "1.2.3")
    assert not version
    assert (
        txt.rstrip()
        == """
# a fist comment line
__hash__ = "4.5.6"
# end of test
__version__ = "1.2.3"
""".rstrip()
    )

    version, txt = code.set_module_var(path, "__version__", "6.7.8")
    assert version == "1.2.3"
    assert (
        txt.rstrip()
        == """
# a fist comment line
__hash__ = "4.5.6"
# end of test
__version__ = "6.7.8"
""".rstrip()
    )

    version, txt = code.set_module_var(path, "__hash__", "9.10.11")
    assert version == "4.5.6"
    assert (
        txt.rstrip()
        == """
# a fist comment line
__hash__ = "9.10.11"
# end of test
__version__ = "6.7.8"
""".rstrip()
    )

    version, txt = code.set_module_var(path, "__version__", "9.10.11")
    assert version == "6.7.8"
    assert (
        txt.rstrip()
        == """
# a fist comment line
__hash__ = "9.10.11"
# end of test
__version__ = "9.10.11"
""".rstrip()
    )


def test_set_module_var_empty_file(tmp_path):
    "check set_module_var creates a new file if not present"
    path = tmp_path / "in1.txt"

    assert not path.exists()
    code.set_module_var(path, "__version__", "1.2.3")

    assert path.exists()
    path.write_text("# a fist comment line\n" + path.read_text().strip())

    code.set_module_var(path, "__hash__", "4.5.6")
    assert (
        path.read_text().strip()
        == """
# a fist comment line
__version__ = "1.2.3"
__hash__ = "4.5.6"
""".strip()
    )
